==> avro-module-2.10.0-20250617-sources.jar: org/mule/weave/v2/module/avro/AvroReader.scala <==
package org.mule.weave.v2.module.avro

import org.apache.avro.AvroRuntimeException
import org.apache.avro.Conversion
import org.apache.avro.Conversions
import org.apache.avro.LogicalType
import org.apache.avro.Schema
import org.apache.avro.Schema.Field
import org.apache.avro.Schema.Type
import org.apache.avro.file.DataFileReader
import org.apache.avro.file.SeekableInput
import org.apache.avro.generic.GenericData.EnumSymbol
import org.apache.avro.generic.GenericDatumReader
import org.apache.avro.generic.GenericRecord
import org.mule.weave.v2.core.io.SeekableStream
import org.mule.weave.v2.model.EvaluationContext
import org.mule.weave.v2.model.capabilities.UnknownLocationCapable
import org.mule.weave.v2.model.structure.AlreadyMaterializedObjectSeq
import org.mule.weave.v2.model.structure.ArraySeq
import org.mule.weave.v2.model.structure.EagerObjectSeq
import org.mule.weave.v2.model.structure.IndexedObjectSeq
import org.mule.weave.v2.model.structure.KeyValuePair
import org.mule.weave.v2.model.structure.ObjectSeq
import org.mule.weave.v2.model.structure.QualifiedName
import org.mule.weave.v2.model.structure.SimpleObjectSeq
import org.mule.weave.v2.model.values.ArrayValue
import org.mule.weave.v2.model.values.KeyValue
import org.mule.weave.v2.model.values.Value
import org.mule.weave.v2.module.DataFormat
import org.mule.weave.v2.module.avro.exception.AvroReadingException
import org.mule.weave.v2.module.option.ConfigurableSchemaSetting
import org.mule.weave.v2.module.pojo.reader.ReflectionJavaValueConverter
import org.mule.weave.v2.module.reader.Reader
import org.mule.weave.v2.module.reader.SourceProvider
import org.mule.weave.v2.module.reader.SourceProviderAwareReader
import org.mule.weave.v2.parser.location.SimpleLocation

import java.io.ByteArrayInputStream
import scala.collection.JavaConverters._

class AvroReader(override val sourceProvider: SourceProvider, val settings: AvroReaderSettings)(implicit ctx: EvaluationContext) extends Reader with SourceProviderAwareReader {

  override def dataFormat: Option[DataFormat[_, _]] = Some(new AvroDataFormat)

  lazy val input = new AvroInput(SeekableStream(sourceProvider.asInputStream))

  override protected def doRead(name: String): Value[_] = {

    // Deserialize users from disk
    val datumReader = {
      settings.schema match {
        case Some(ConfigurableSchemaSetting.SchemaResult(schemaResult)) =>
          val schema = new Schema.Parser().parse(new ByteArrayInputStream(schemaResult))
          new GenericDatumReader[GenericRecord](schema)
        case _ =>
          new GenericDatumReader[GenericRecord]()
      }
    }
    val dataFileReader: DataFileReader[GenericRecord] = new DataFileReader[GenericRecord](input, datumReader)
    val scala: Iterator[GenericRecord] = dataFileReader.iterator().asScala
    ArrayValue(
      ArraySeq(scala.zipWithIndex.map((gr) => {
        new AvroObjectValue(new AvroObjectSeq(gr._1), () => s"root(${gr._2})")
      }), materializedValues = true),
      UnknownLocationCapable)
  }
}

class AvroObjectSeq(record: GenericRecord) extends EagerObjectSeq with IndexedObjectSeq with SimpleObjectSeq with AlreadyMaterializedObjectSeq {

  def calculateValueRealSchema(schema: Schema, value: Any): Option[Schema] = {
    val theType = schema.getType
    if (theType == Type.NULL) {
      if (value == null) {
        Some(schema)
      } else {
        None
      }
    } else if (theType == Type.ARRAY) {
      if (value.getClass.isArray) {
        Some(schema)
      } else {
        None
      }
    } else if (schema.isUnion) {
      val option = schema.getTypes.asScala.toStream
        .flatMap((s) => calculateValueRealSchema(s, value))
        .headOption
      option
    } else if (theType == Type.ENUM) {
      value match {
        case es: EnumSymbol if (schema.getEnumSymbols.contains(es.toString)) =>
          Some(schema)
        case str: String if (schema.getEnumSymbols.contains(str)) =>
          Some(schema)
        case _ =>
          None
      }
    } else {
      val expectedClass = ConversionFactory.typeToClassMap.get(theType)
      expectedClass match {
        case Some(theClass) => {
          if (theClass.isInstance(value)) {
            Some(schema)
          } else {
            None
          }
        }
        case None => {
          Some(schema)
        }
      }
    }
  }

  private def toKeyValuePair(field: Field)(implicit ctx: EvaluationContext) = {
    val avroValue: Value[_] = toAvroValue(field)
    KeyValuePair(KeyValue(field.name()), avroValue)
  }

  private def toAvroValue(field: Field)(implicit ctx: EvaluationContext): Value[_] = {
    val value: AnyRef = record.get(field.pos())
    val realSchema = if (field.schema().isUnion) {
      val maybeSchema = calculateValueRealSchema(field.schema(), value)
      maybeSchema.getOrElse(field.schema())
    } else {
      field.schema()
    }
    val logicalType: LogicalType = realSchema.getLogicalType
    val maybeConverter: Option[Conversion[_]] = ConversionFactory.getConversion(logicalType)
    val convertedValue = maybeConverter match {
      case Some(converter) => {
        try {
          Conversions.convertToLogicalType(value, realSchema, logicalType, converter)
        } catch {
          case ae: AvroRuntimeException     => throw new AvroReadingException(ae.getMessage, SimpleLocation(field.name()))
          case ae: IllegalArgumentException => throw new AvroReadingException(ae.getMessage, SimpleLocation(field.name()))
        }
      }
      case None => value
    }
    ReflectionJavaValueConverter.convert(convertedValue, () => field.name())
  }

  override def apply(index: Long)(implicit ctx: EvaluationContext): KeyValuePair = {
    val fields = record.getSchema.getFields
    if (index >= 0 && index < size()) {
      val field: Field = fields.get(index.toInt)
      toKeyValuePair(field)
    } else {
      null
    }
  }

  override def selectKeyValue(key: Value[QualifiedName])(implicit ctx: EvaluationContext): KeyValuePair = {
    val name = key.evaluate.name
    val field: Field = record.getSchema.getField(name)
    if (field != null) {
      toKeyValuePair(field)
    } else {
      null
    }
  }

  override def selectValue(key: Value[QualifiedName])(implicit ctx: EvaluationContext): Value[_] = {
    val name = key.evaluate.name
    val field = record.getSchema.getField(name)
    if (field != null) {
      toKeyValuePair(field)._2
    } else {
      null
    }
  }

  override def allKeyValuesOf(key: Value[QualifiedName])(implicit ctx: EvaluationContext): Option[ObjectSeq] = {
    keyValueOf(key).map((kvp) => ObjectSeq(kvp))
  }

  override def size()(implicit ctx: EvaluationContext): Long = {
    record.getSchema.getFields.size()
  }

  override def isEmpty()(implicit ctx: EvaluationContext): Boolean = {
    record.getSchema.getFields.isEmpty
  }

  override def keyValueOfWithIndex(key: Value[QualifiedName])(implicit ctx: EvaluationContext): IndexedSelection[KeyValuePair] = {
    val name = key.evaluate.name
    val field = record.getSchema.getField(name)
    if (field != null) {
      IndexedSelection(field.pos(), toKeyValuePair(field))
    } else {
      null
    }
  }

  override def selectValueWithIndex(key: Value[QualifiedName])(implicit ctx: EvaluationContext): IndexedSelection[Value[_]] = {
    val name = key.evaluate.name
    val field = record.getSchema.getField(name)
    if (field != null) {
      IndexedSelection(field.pos(), toAvroValue(field))
    } else {
      null
    }
  }
}

class AvroInput(seekableStream: SeekableStream) extends SeekableInput {

  override def seek(p: Long): Unit = seekableStream.seek(p)

  override def tell(): Long = seekableStream.position()

  override def length(): Long = seekableStream.size()

  override def read(b: Array[Byte], off: Int, len: Int): Int = seekableStream.read(b, off, len)

  override def close(): Unit = seekableStream.close()
}

==> avro-module-2.10.1-20250903-sources.jar: org/mule/weave/v2/module/avro/AvroReader.scala <==
package org.mule.weave.v2.module.avro

import org.apache.avro.AvroRuntimeException
import org.apache.avro.Conversion
import org.apache.avro.Conversions
import org.apache.avro.LogicalType
import org.apache.avro.Schema
import org.apache.avro.Schema.Field
import org.apache.avro.Schema.Type
import org.apache.avro.file.DataFileReader
import org.apache.avro.file.SeekableInput
import org.apache.avro.generic.GenericData
import org.apache.avro.generic.GenericData.EnumSymbol
import org.apache.avro.generic.GenericDatumReader
import org.apache.avro.generic.GenericRecord
import org.mule.weave.v2.core.io.SeekableStream
import org.mule.weave.v2.model.EvaluationContext
import org.mule.weave.v2.model.capabilities.UnknownLocationCapable
import org.mule.weave.v2.model.structure.AlreadyMaterializedObjectSeq
import org.mule.weave.v2.model.structure.ArraySeq
import org.mule.weave.v2.model.structure.EagerObjectSeq
import org.mule.weave.v2.model.structure.IndexedObjectSeq
import org.mule.weave.v2.model.structure.KeyValuePair
import org.mule.weave.v2.model.structure.ObjectSeq
import org.mule.weave.v2.model.structure.QualifiedName
import org.mule.weave.v2.model.structure.SimpleObjectSeq
import org.mule.weave.v2.model.values.ArrayValue
import org.mule.weave.v2.model.values.KeyValue
import org.mule.weave.v2.model.values.Value
import org.mule.weave.v2.module.DataFormat
import org.mule.weave.v2.module.avro.AvroReader.convertToLogicalType
import org.mule.weave.v2.module.avro.exception.AvroReadingException
import org.mule.weave.v2.module.option.ConfigurableSchemaSetting
import org.mule.weave.v2.module.pojo.reader.ReflectionJavaValueConverter
import org.mule.weave.v2.module.reader.Reader
import org.mule.weave.v2.module.reader.SourceProvider
import org.mule.weave.v2.module.reader.SourceProviderAwareReader
import org.mule.weave.v2.parser.location.SimpleLocation

import java.io.ByteArrayInputStream
import scala.collection.JavaConverters._

class AvroReader(override val sourceProvider: SourceProvider, val settings: AvroReaderSettings)(implicit ctx: EvaluationContext) extends Reader with SourceProviderAwareReader {

  override def dataFormat: Option[DataFormat[_, _]] = Some(new AvroDataFormat)

  lazy val input = new AvroInput(SeekableStream(sourceProvider.asInputStream))

  override protected def doRead(name: String): Value[_] = {

    val (schema, dataFileReader) = {
      settings.schema match {
        case Some(ConfigurableSchemaSetting.SchemaResult(schemaResult)) => {
          val schema = new Schema.Parser().parse(new ByteArrayInputStream(schemaResult))
          val datumReader = createDatumReader(schema)
          val reader = new DataFileReader[Any](input, datumReader.asInstanceOf[GenericDatumReader[Any]])
          (schema, reader)
        }
        case _ => {
          val genericReader = new DataFileReader[Any](input, new GenericDatumReader[Any]())
          val schema = genericReader.getSchema
          (schema, genericReader)
        }
      }
    }

    val scala: Iterator[Any] = dataFileReader.iterator().asScala

    schema.getType match {
      case Type.RECORD => {
        ArrayValue(
          ArraySeq(scala.zipWithIndex.map((valueWithIndex) => {
            val value = valueWithIndex._1
            val index = valueWithIndex._2
            value match {
              case gr: GenericRecord => new AvroObjectValue(new AvroObjectSeq(gr), () => s"root(${index})")
              case _                 => throw new AvroReadingException(s"Expected GenericRecord but got ${value.getClass}", SimpleLocation(name))
            }
          }), materializedValues = true),
          UnknownLocationCapable)
      }

      case Type.ARRAY => {
        ArrayValue(
          ArraySeq(scala.zipWithIndex.map((valueWithIndex) => {
            val value = valueWithIndex._1
            val index = valueWithIndex._2
            value match {
              case array: GenericData.Array[_] => ReflectionJavaValueConverter.convert(array, () => s"root(${index})")
              case _                           => throw new AvroReadingException(s"Expected GenericData.Array but got ${value.getClass}", SimpleLocation(name))
            }
          }), materializedValues = true),
          UnknownLocationCapable)
      }

      case Type.NULL | Type.BOOLEAN | Type.INT | Type.LONG | Type.FLOAT | Type.DOUBLE | Type.STRING | Type.BYTES => {
        ArrayValue(
          ArraySeq(scala.zipWithIndex.map((valueWithIndex) => {
            val value = valueWithIndex._1
            val index = valueWithIndex._2
            val location = s"root(${index})"
            val convertedValue = convertToLogicalType(value, schema, location)
            ReflectionJavaValueConverter.convert(convertedValue, () => location)
          }), materializedValues = true),
          UnknownLocationCapable)
      }

      case invalidType => {
        throw new AvroReadingException(s"Unsupported top-level schema type: ${invalidType}", SimpleLocation(name))
      }
    }
  }

  private def createDatumReader(schema: Schema): GenericDatumReader[_] = {
    schema.getType match {
      case Type.RECORD => new GenericDatumReader[GenericRecord](schema)
      case Type.ARRAY  => new GenericDatumReader[GenericData.Array[_]](schema)
      case Type.MAP    => new GenericDatumReader[java.util.Map[String, Any]](schema)
      case Type.UNION  => new GenericDatumReader[Any](schema)
      case Type.ENUM   => new GenericDatumReader[EnumSymbol](schema)
      case Type.FIXED  => new GenericDatumReader[GenericData.Fixed](schema)
      case Type.NULL | Type.BOOLEAN | Type.INT | Type.LONG | Type.FLOAT | Type.DOUBLE | Type.STRING | Type.BYTES =>
        new GenericDatumReader[Any](schema)
      case invalidType =>
        throw new AvroReadingException(s"Unsupported schema type: ${invalidType}", SimpleLocation("schema"))
    }
  }
}

object AvroReader {
  def convertToLogicalType(value: Any, schema: Schema, location: String): Any = {
    val logicalType: LogicalType = schema.getLogicalType
    val maybeConverter: Option[Conversion[_]] = ConversionFactory.getConversion(logicalType)
    val convertedValue = maybeConverter match {
      case Some(converter) => {
        try {
          Conversions.convertToLogicalType(value, schema, logicalType, converter)
        } catch {
          case ae: AvroRuntimeException =>
            throw new AvroReadingException(ae.getMessage, SimpleLocation(location))
          case ae: IllegalArgumentException => throw new AvroReadingException(ae.getMessage, SimpleLocation(location))
        }
      }
      case None => value
    }
    convertedValue
  }
}

class AvroObjectSeq(record: GenericRecord) extends EagerObjectSeq with IndexedObjectSeq with SimpleObjectSeq with AlreadyMaterializedObjectSeq {

  def calculateValueRealSchema(schema: Schema, value: Any): Option[Schema] = {
    val theType = schema.getType
    if (theType == Type.NULL) {
      if (value == null) {
        Some(schema)
      } else {
        None
      }
    } else if (theType == Type.ARRAY) {
      if (value.getClass.isArray) {
        Some(schema)
      } else {
        None
      }
    } else if (schema.isUnion) {
      val option = schema.getTypes.asScala.toStream
        .flatMap((s) => calculateValueRealSchema(s, value))
        .headOption
      option
    } else if (theType == Type.ENUM) {
      value match {
        case es: EnumSymbol if (schema.getEnumSymbols.contains(es.toString)) =>
          Some(schema)
        case str: String if (schema.getEnumSymbols.contains(str)) =>
          Some(schema)
        case _ =>
          None
      }
    } else {
      val expectedClass = ConversionFactory.typeToClassMap.get(theType)
      expectedClass match {
        case Some(theClass) => {
          if (theClass.isInstance(value)) {
            Some(schema)
          } else {
            None
          }
        }
        case None => {
          Some(schema)
        }
      }
    }
  }

  private def toKeyValuePair(field: Field)(implicit ctx: EvaluationContext) = {
    val avroValue: Value[_] = toAvroValue(field)
    KeyValuePair(KeyValue(field.name()), avroValue)
  }

  private def toAvroValue(field: Field)(implicit ctx: EvaluationContext): Value[_] = {
    val value: AnyRef = record.get(field.pos())
    val realSchema = if (field.schema().isUnion) {
      val maybeSchema = calculateValueRealSchema(field.schema(), value)
      maybeSchema.getOrElse(field.schema())
    } else {
      field.schema()
    }
    val location = field.name()

    val convertedValue: Any = convertToLogicalType(value, realSchema, location)
    ReflectionJavaValueConverter.convert(convertedValue, () => field.name())
  }

  override def apply(index: Long)(implicit ctx: EvaluationContext): KeyValuePair = {
    val fields = record.getSchema.getFields
    if (index >= 0 && index < size()) {
      val field: Field = fields.get(index.toInt)
      toKeyValuePair(field)
    } else {
      null
    }
  }

  override def selectKeyValue(key: Value[QualifiedName])(implicit ctx: EvaluationContext): KeyValuePair = {
    val name = key.evaluate.name
    val field: Field = record.getSchema.getField(name)
    if (field != null) {
      toKeyValuePair(field)
    } else {
      null
    }
  }

  override def selectValue(key: Value[QualifiedName])(implicit ctx: EvaluationContext): Value[_] = {
    val name = key.evaluate.name
    val field = record.getSchema.getField(name)
    if (field != null) {
      toKeyValuePair(field)._2
    } else {
      null
    }
  }

  override def allKeyValuesOf(key: Value[QualifiedName])(implicit ctx: EvaluationContext): Option[ObjectSeq] = {
    keyValueOf(key).map((kvp) => ObjectSeq(kvp))
  }

  override def size()(implicit ctx: EvaluationContext): Long = {
    record.getSchema.getFields.size()
  }

  override def isEmpty()(implicit ctx: EvaluationContext): Boolean = {
    record.getSchema.getFields.isEmpty
  }

  override def keyValueOfWithIndex(key: Value[QualifiedName])(implicit ctx: EvaluationContext): IndexedSelection[KeyValuePair] = {
    val name = key.evaluate.name
    val field = record.getSchema.getField(name)
    if (field != null) {
      IndexedSelection(field.pos(), toKeyValuePair(field))
    } else {
      null
    }
  }

  override def selectValueWithIndex(key: Value[QualifiedName])(implicit ctx: EvaluationContext): IndexedSelection[Value[_]] = {
    val name = key.evaluate.name
    val field = record.getSchema.getField(name)
    if (field != null) {
      IndexedSelection(field.pos(), toAvroValue(field))
    } else {
      null
    }
  }
}

class AvroInput(seekableStream: SeekableStream) extends SeekableInput {

  override def seek(p: Long): Unit = seekableStream.seek(p)

  override def tell(): Long = seekableStream.position()

  override def length(): Long = seekableStream.size()

  override def read(b: Array[Byte], off: Int, len: Int): Int = seekableStream.read(b, off, len)

  override def close(): Unit = seekableStream.close()
}
